from .auxiliary import LacssCollaborator
from .convnext import ConvNeXt
from .detector import Detector
from .lacss import Lacss
from .lpn import LPN
from .resnet import ResNet
from .segmentor import Segmentor
from .unet import UNet

__all__ = [
    "LacssCollaborator",
    "ConvNeXt",
    "Detector",
    "Lacss",
    "LPN",
    "ResNet",
    "Segmentor",
    "UNet",
]



import jax
import jax.numpy as jnp
import numpy as np


def select_sample(chains):
    """a naive sample selection logic
    find the mean tracked_vector, and return the sample that has the most similar values
    """

    a = chains["tracked"] - chains["tracked"].mean(axis=1, keepdims=True)
    a = (a**2).sum(axis=(0, 2))

    return np.argmin(a)


def update_with_sample(df, chains, idx):

    tracked = chains["tracked"][:, idx, :]
    parent = chains["parent"][:, idx, :]
    age = chains["age"][:, idx, :]

    df_tracking = df.copy()
    df_tracking["parent"] = -1
    df_tracking["tracked"] = False
    df_tracking["age"] = 0

    n_frame = len(tracked)
    ndets = [
        len(df_tracking.loc[df_tracking["frame"] == k + 1]) for k in range(n_frame)
    ]
    first_ids = [
        df_tracking.loc[df_tracking["frame"] == k + 1, "index"].iloc[0]
        for k in range(n_frame)
    ]

    def pad_or_truncate(a, length, constant_values=0):
        if len(a) >= length:
            return a[:length]
        else:
            return np.pad(a, [0, length - len(a)], constant_values=constant_values)

    for k in range(n_frame, 0, -1):
        t = pad_or_truncate(
            tracked[k - 1],
            ndets[k - 1],
        )
        df_tracking.loc[df_tracking["frame"] == k, "tracked"] = t

        p = pad_or_truncate(parent[k - 1], ndets[k - 1], constant_values=-1)
        p = np.where(p >= 0, p + first_ids[k - 2], p)
        df_tracking.loc[df_tracking["frame"] == k, "parent"] = p

        a = pad_or_truncate(
            age[k - 1],
            ndets[k - 1],
        )
        df_tracking.loc[df_tracking["frame"] == k, "age"] = a

    df_tracking = df_tracking.loc[df_tracking["tracked"]].drop("tracked", axis=1)

    df_tracking["cell_id"] = -1
    cell_id = 0

    for id0 in df_tracking.index[::-1]:
        if df_tracking.loc[id0, "cell_id"] != -1:
            continue

        id1 = id0
        while id1 != -1:
            df_tracking.loc[id1, "cell_id"] = cell_id
            if df_tracking.loc[id1, "age"] != 0:
                id1 = df_tracking.loc[id1, "parent"]
            else:
                break
        cell_id += 1

    df_tracking["cell_id"] = df_tracking["cell_id"].max() - df_tracking["cell_id"]

    df_tracking["child_1"] = -1
    df_tracking["child_2"] = -1
    df_tracking["parent_cell_id"] = -1

    for cell_id in range(df_tracking["cell_id"].max() + 1):

        parent_id = df_tracking.loc[
            df_tracking["cell_id"] == cell_id, "parent"
        ].to_numpy()[0]

        if parent_id != -1:
            parent_cell_id = df_tracking.loc[parent_id, "cell_id"]

            df_tracking.loc[
                df_tracking["cell_id"] == cell_id, "parent_cell_id"
            ] = parent_cell_id

            if (
                df_tracking.loc[
                    df_tracking["cell_id"] == parent_cell_id, "child_1"
                ].iloc[0]
                == -1
            ):
                df_tracking.loc[
                    df_tracking["cell_id"] == parent_cell_id, "child_1"
                ] = cell_id
            else:
                df_tracking.loc[
                    df_tracking["cell_id"] == parent_cell_id, "child_2"
                ] = cell_id

    df_tracking = df_tracking.rename(columns={"parent": "parent_id"})
    df_tracking = df_tracking.rename(columns={"parent_cell_id": "parent"})

    return df_tracking


def _box_intersect(gt, pred):
    start_frame = max(gt["frame"].min(), pred["frame"].min())
    end_frame = min(gt["frame"].max(), pred["frame"].max())

    gt = gt.loc[(gt["frame"] >= start_frame) & (gt["frame"] <= end_frame)]
    pred = pred.loc[(pred["frame"] >= start_frame) & (pred["frame"] <= end_frame)]

    gt_y0, gt_x0, gt_h, gt_w = (
        gt[["bbox_y", "bbox_x", "bbox_h", "bbox_w"]].to_numpy().transpose()
    )
    gt_y1 = gt_y0 + gt_h
    gt_x1 = gt_x0 + gt_w

    pred_y0, pred_x0, pred_h, pred_w = (
        pred[["bbox_y", "bbox_x", "bbox_h", "bbox_w"]].to_numpy().transpose()
    )
    pred_y1 = pred_y0 + pred_h
    pred_x1 = pred_x0 + pred_w

    y_min_max = np.minimum(gt_y1, pred_y1)
    y_max_min = np.maximum(gt_y0, pred_y0)
    x_min_max = np.minimum(gt_x1, pred_x1)
    x_max_min = np.maximum(gt_x0, pred_x0)

    intersect_heights = y_min_max - y_max_min
    intersect_widths = x_min_max - x_max_min
    intersect_heights = np.maximum(0, intersect_heights)
    intersect_widths = np.maximum(0, intersect_widths)

    return intersect_heights * intersect_widths


def broi(gt, pred):
    gt_area = (gt["bbox_h"] * gt["bbox_w"]).sum()
    pred_area = (pred["bbox_h"] * pred["bbox_w"]).sum()
    intersects = _box_intersect(gt, pred).sum()
    roi = intersects / (gt_area + pred_area - intersects + 1e-8)
    return roi


def proi(gt, pred, g=30):
    total_len = len(gt) + len(pred)

    start_frame = max(gt["frame"].min(), pred["frame"].min())
    end_frame = min(gt["frame"].max(), pred["frame"].max())

    gt = gt.loc[(gt["frame"] >= start_frame) & (gt["frame"] <= end_frame)]
    pred = pred.loc[(pred["frame"] >= start_frame) & (pred["frame"] <= end_frame)]

    yofs = np.minimum(np.abs(gt["y"].to_numpy() - pred["y"].to_numpy()) / g, 1.0)
    xofs = np.minimum(np.abs(gt["x"].to_numpy() - pred["x"].to_numpy()) / g, 1.0)
    its = ((1 - yofs) * (1 - xofs)).sum()
    proi = its / (total_len - its + 1e-8)
    return proi


def filter_ious(ious):

    ious_out = np.zeros_like(ious)
    ious = ious.copy()

    for _ in range(len(ious)):
        row, col = np.unravel_index(ious.argmax(), ious.shape)
        ious_out[row, col] = ious[row, col]
        ious[row, :] = -1
        ious[:, col] = -1

    return ious_out



from __future__ import annotations

from typing import Mapping, Optional, Sequence, Tuple

import flax.linen as nn
import jax
import jax.numpy as jnp

from ..ops import locations_to_labels
from ..typing import *


class LPN(nn.Module):
    """Location detection head

    Attributes:

        feature_levels: Input feature level, e.g. [2, 3, 4]
        conv_spec: Conv layer specification
        detection_roi: Parameter for label smoothing

    """

    feature_levels: Sequence[int] = (4, 3, 2)
    conv_spec: Tuple[Sequence[int], Sequence[int]] = ((384, 384, 384, 384), ())
    detection_roi: float = 8.0

    @nn.compact
    def _process_feature(
        self, feature: ArrayLike
    ) -> Tuple[ArrayLike, ArrayLike, ArrayLike]:
        conv_spec = self.conv_spec

        x = feature

        for n_ch in conv_spec[0]:
            x = nn.Conv(n_ch, (3, 3), use_bias=False)(x)
            x = nn.GroupNorm(num_groups=n_ch)(x[None, ...])[0]
            x = jax.nn.relu(x)

        for n_ch in conv_spec[1]:
            x = nn.Conv(n_ch, (1, 1), use_bias=False)(x)
            x = nn.GroupNorm(num_groups=n_ch)(x[None, ...])[0]
            x = jax.nn.relu(x)

        scores_out = nn.Conv(1, (1, 1))(x)
        scores_out = jax.nn.sigmoid(scores_out)

        regression_out = nn.Conv(2, (1, 1))(x)

        return scores_out, regression_out, x

    def __call__(
        self,
        inputs: Mapping[str, ArrayLike],
        scaled_gt_locations: Optional[ArrayLike] = None,
    ) -> dict:
        """
        Args:
            inputs: feature dict: {'lvl': [H, W, C]}
            scaled_gt_locations: scaled 0..1 [N, 2], only valid in training

        Returns:
            A dict of features

                * lpn_scores: dict: {'lvl': [H, W, 1]}
                * lpn_regressions: dict {'lvl': [H, W, 2]}
                * gt_lpn_scores: dict {'lvl': [H, W, 1]}, only if training
                * gt_lpn_regressions: dict {'lvl': [H, W, 2]}, only if training

        """
        all_scores = dict()
        all_regrs = dict()
        all_features = dict()

        for lvl in self.feature_levels:

            feature = inputs[str(lvl)]
            score, regression, lpn_feature = self._process_feature(feature)
            all_scores[str(lvl)] = score
            all_regrs[str(lvl)] = regression
            all_features[str(lvl)] = lpn_feature

        outputs = dict(
            lpn_features=all_features,
            lpn_scores=all_scores,
            lpn_regressions=all_regrs,
        )

        if scaled_gt_locations is not None:

            all_gt_scores = dict()
            all_gt_regrs = dict()

            for lvl in self.feature_levels:

                feature_shape = inputs[str(lvl)].shape[-3:-1]

                score_target, regression_target = locations_to_labels(
                    scaled_gt_locations,
                    target_shape=feature_shape,
                    threshold=self.detection_roi / (2**lvl),
                )
                all_gt_scores[str(lvl)] = score_target
                all_gt_regrs[str(lvl)] = regression_target

            outputs.update(
                dict(
                    lpn_gt_scores=all_gt_scores,
                    lpn_gt_regressions=all_gt_regrs,
                )
            )

        return outputs



from __future__ import annotations

from functools import partial

import jax
import jax.numpy as jnp
import optax

from ..ops import sorbel_edges
from ..utils import unpack_x_y_sample_weight
from .common import binary_focal_factor_loss

EPS = jnp.finfo("float32").eps


def _compute_edge(instance_output, instance_yc, instance_xc, height, width):
    _, ps, _ = instance_yc.shape
    padding = ps // 2 + 1

    patch_edges = jnp.square(sorbel_edges(instance_output))
    patch_edges = (patch_edges[0] + patch_edges[1]) / 8.0
    patch_edges = jnp.sqrt(jnp.clip(patch_edges, 1e-8, 1.0))  # avoid GPU error
    # patch_edges = jnp.where(patch_edges > 0, jnp.sqrt(patch_edges), 0)
    combined_edges = jnp.zeros([height + padding * 2, width + padding * 2])
    combined_edges = combined_edges.at[
        instance_yc + padding, instance_xc + padding
    ].add(patch_edges)
    combined_edges = combined_edges[
        padding : padding + height, padding : padding + width
    ]
    combined_edges = jnp.tanh(combined_edges)

    return combined_edges


def self_supervised_edge_loss(batch, prediction):
    """Cell border prediction consistency loss"""

    preds = prediction
    inputs, _, _ = unpack_x_y_sample_weight(batch)

    instance_output = preds["instance_output"]
    instance_yc = preds["instance_yc"]
    instance_xc = preds["instance_xc"]
    height, width = inputs["image"].shape[:2]
    instance_edge = _compute_edge(
        instance_output, instance_yc, instance_xc, height, width
    )
    instance_edge_pred = jax.nn.sigmoid(preds["edge_pred"])

    return optax.l2_loss(instance_edge_pred, instance_edge).mean()


def self_supervised_segmentation_loss(
    batch, prediction, *, offset_sigma=jnp.array([10.0]), offset_scale=jnp.array([2.0])
):
    """Image segmentation consistenct loss for the collaboraor model"""

    preds = prediction
    inputs, _, _ = unpack_x_y_sample_weight(batch)

    height, width = inputs["image"].shape[:2]
    ps = preds["instance_yc"].shape[-1]

    offset_sigma = jnp.asarray(offset_sigma).reshape(-1)
    offset_scale = jnp.asarray(offset_scale).reshape(-1)

    def _max_merge(pred):
        label = jnp.zeros([height + ps, width + ps]) - 1.0e6
        yc, xc = pred["instance_yc"], pred["instance_xc"]
        logit = pred["instance_logit"]

        if (
            offset_sigma.size > 1
            and "cls_id" in inputs
            and inputs["cls_id"] is not None
        ):
            c = inputs["cls_id"].astype(int).squeeze()
        else:
            c = 0

        sigma = offset_sigma[c]
        scale = offset_scale[c]
        offset = (jnp.mgrid[:ps, :ps] - (ps - 1) / 2) ** 2 / (2 * sigma * sigma)
        offset = jnp.exp(-offset.sum(axis=0)) * scale
        logit += offset

        label = label.at[yc + ps // 2, xc + ps // 2].max(logit)
        label = label[ps // 2 : -ps // 2, ps // 2 : -ps // 2]

        return label

    # fg_pred = jax.nn.tanh(preds["fg_pred"] / 2)
    # fg_label = jax.nn.tanh(_max_merge(preds) / 2)
    fg_pred = jax.nn.sigmoid(preds["fg_pred"])
    fg_label = jax.lax.stop_gradient(jax.nn.sigmoid(_max_merge(preds)))

    # locs = jnp.floor(inputs["gt_locations"] + 0.5)
    # locs = locs.astype(int)
    # fg_label = fg_label.at[locs[:, 0], locs[:, 1]].set(1.0)

    assert fg_pred.shape == fg_label.shape

    # loss = (1.0 - fg_label) * fg_pred + fg_label * (1.0 - fg_pred)
    # p_t = fg_label * fg_pred + (1.0 -fg_label) * (1.0 - fg_pred)
    # loss = -jnp.log(jnp.clip(p_t, EPS, 1))
    loss = binary_focal_factor_loss(fg_pred, fg_label, gamma=1)
    loss = loss.mean()

    return loss


def aux_size_loss(batch, prediction, *, weight=0.01):
    """Auxillary loss to prevent model collapse"""
    preds = prediction
    inputs, labels, _ = unpack_x_y_sample_weight(batch)

    if labels is None:
        labels = {}

    if "gt_labels" in labels or "gt_masks" in labels:
        return 0.0

    height, width = inputs["image"].shape[:2]

    valid_locs = (
        (preds["instance_yc"] >= 0)
        & (preds["instance_yc"] < height)
        & (preds["instance_xc"] >= 0)
        & (preds["instance_xc"] < width)
    )

    areas = jnp.sum(
        preds["instance_output"], axis=(-1, -2), where=valid_locs, keepdims=True
    )
    areas = jnp.clip(areas, EPS, 1e6)
    loss = jax.lax.rsqrt(areas) * preds["instance_output"].shape[-1]

    mask = preds["instance_mask"]
    n_instances = jnp.count_nonzero(mask) + EPS
    loss = loss.sum(where=mask) / n_instances

    return loss * weight


def supervised_segmentation_loss(batch, prediction):
    preds = prediction
    _, labels, _ = unpack_x_y_sample_weight(batch)

    if "gt_labels" in labels:

        mask = (labels["gt_labels"] > 0).astype("float32")

    else:

        mask = labels["gt_image_mask"].astype("float32")

    return optax.sigmoid_binary_cross_entropy(preds["fg_pred"], mask).mean()


def collaborator_segm_loss(batch, prediction, *, sigma, pi):
    preds = prediction
    inputs, labels, _ = unpack_x_y_sample_weight(batch)

    if not "fg_pred" in preds:
        return 0.0

    if labels is None:
        labels = {}

    if "gt_image_mask" in labels or "gt_labels" in labels:
        return supervised_segmentation_loss(batch, prediction)
    else:
        return self_supervised_segmentation_loss(
            batch, prediction, offset_sigma=sigma, offset_scale=pi
        )


def collaborator_border_loss(batch, prediction):
    preds = prediction
    _, labels, _ = unpack_x_y_sample_weight(batch)

    if not "edge_pred" in preds:
        return 0.0

    if labels is None:
        labels = {}

    if "gt_labels" in labels or "gt_masks" in labels:
        return 0.0

    else:
        return self_supervised_edge_loss(batch, prediction)



from __future__ import annotations

from pathlib import Path
from typing import Any, Mapping, Protocol, TypedDict, Union

import numpy as np
import optax
from flax.core.frozen_dict import FrozenDict
from jax import Array
from jax.typing import ArrayLike

Params = FrozenDict

Optimizer = optax.GradientTransformation

PathLike = Union[str, Path]


class LossFunc(Protocol):
    def __call__(self, batch: Any, prediction: Any) -> float:
        ...


class Metric(Protocol):
    def update(self, batch: Any, prediction: Any):
        ...

    def compute(self, *args, **kwargs) -> dict:
        ...


class Patches(TypedDict):
    instance_output: ArrayLike
    instance_yc: ArrayLike
    instance_xc: ArrayLike


DataDict = Mapping[str, ArrayLike]

__all__ = [
    "Array",
    "ArrayLike",
    "DataDict",
    "Params",
    "Optimizer",
    "LossFunc",
    "Metric",
    "Patches",
    "PathLike",
]



from __future__ import annotations

import math
from functools import partial
from typing import Mapping, Sequence

import flax.linen as nn
import jax
import jax.numpy as jnp

from lacss.ops import gather_patches
from lacss.typing import *

from ..typing import *
from .common import SpatialAttention


class _Encoder(nn.Module):
    n_output_features: int
    input_patch_size: int
    encoding_dims: tuple[int, int, int] = (8, 8, 4)
    n_latent_features: int = -1

    @nn.compact
    def __call__(self, feature: ArrayLike, loc: ArrayLike) -> Array:
        patch_center, _, _, _ = gather_patches(feature, loc, patch_size=2)
        encodings = patch_center.mean(axis=(-2, -3))

        dim = self.n_latent_features
        dim = dim if dim > 0 else math.prod(self.encoding_dims)
        encodings = nn.Dense(dim)(encodings)
        encodings = jax.nn.relu(encodings)
        encodings = nn.Dense(dim)(encodings)
        encodings = jax.nn.relu(encodings)

        new_dim = math.prod(self.encoding_dims)
        encoding_shape = (-1,) + self.encoding_dims
        encodings = nn.Dense(new_dim)(encodings).reshape(encoding_shape)
        encodings = jax.image.resize(
            encodings,
            (
                encodings.shape[0],
                self.input_patch_size,
                self.input_patch_size,
                self.encoding_dims[-1],
            ),
            "linear",
        )

        encodings = nn.Conv(self.n_output_features, (1, 1), use_bias=False)(encodings)

        return encodings


class Segmentor(nn.Module):
    """LACSS segmentation head.

    Attributes:
        feature_level: The scale of the feature used for segmentation
        conv_spec: conv_block definition, e.g. ((384,384,384), (64,))
        instance_crop_size: Crop size for segmentation.
        with_attention: Whether use spatial attention layer.
        learned_encoding: Whether to use hard-coded position encoding.
        encoder_dims: Dim of the position encoder, if using learned encoding. Default is (8,8,4)

    """

    feature_level: int = 2
    conv_spec: tuple[Sequence[int], Sequence[int]] = (
        (384, 384, 384),
        (64,),
    )
    instance_crop_size: int = 96
    use_attention: bool = False
    learned_encoding: bool = True
    encoder_dims: Sequence[int] = (8, 8, 4)
    # n_cls: int = -1

    # if not feature_level in (0,1,2):
    #     raise ValueError('feature_level should be 1,2 or 0')

    @nn.compact
    def __call__(
        self, features: Mapping[str, ArrayLike], locations: ArrayLike
    ) -> DataDict:
        """
        Args:
            features: {'scale' [H, W, C]} feature dictionary from the backbone.
            locations: [N, 2]  normalized to image size.

        Returns:
            A dictionary of values representing segmentation outputs.

                * instance_output: [N, crop_size, crop_size]
                * instance_mask; [N, 1, 1] boolean mask indicating valid outputs
                * instance_yc: [N, crop_size, crop_size] meshgrid y coordinates
                * instance_xc: [N, crop_size, crop_size] meshgrid x coordinates
        """
        crop_size = self.instance_crop_size
        lvl = self.feature_level
        scale = 2**lvl
        conv_spec = self.conv_spec
        patch_size = crop_size // scale

        x = features[str(lvl)]
        # h, w, _ = x.shape

        # feature convs
        for n_ch in conv_spec[0]:
            x = nn.Conv(n_ch, (3, 3), use_bias=False)(x)
            x = nn.GroupNorm(num_groups=n_ch)(x[None, ...])[0]
            x = jax.nn.relu(x)

        # mixing features with pos_encodings
        n_ch = conv_spec[1][0]
        x = nn.Conv(n_ch, (3, 3), use_bias=False)(x)
        patches, y0, x0, _ = gather_patches(x, locations, patch_size=patch_size)

        if not self.learned_encoding:
            encodings = (
                jnp.mgrid[:patch_size, :patch_size] / patch_size - 0.5
            ).transpose(1, 2, 0)
            encodings = nn.Conv(n_ch, (1, 1), use_bias=False)(encodings)
        else:
            encodings = _Encoder(
                n_ch,
                patch_size,
                self.encoder_dims,
            )(features[str(lvl)], locations)

        patches += encodings
        mask = jnp.expand_dims((locations >= 0).any(axis=-1), (1, 2))
        patches = jax.nn.relu(patches)

        if self.use_attention:
            patches = SpatialAttention()(patches)

        # patche convs
        for n_ch in conv_spec[1][1:]:
            patches = nn.Conv(n_ch, (3, 3))(patches)
            patches = jax.nn.relu(patches)

        # outputs
        if scale == 1:
            logits = nn.Conv(1, (3, 3))(patches)
        else:
            logits = nn.ConvTranspose(1, (2, 2), strides=(2, 2))(patches)
            if scale == 4:
                logits = jax.image.resize(
                    logits, patches.shape[:1] + (crop_size, crop_size, 1), "linear"
                )

        logits = logits.squeeze(-1)
        outputs = jax.nn.sigmoid(logits)

        # indicies
        yc, xc = jnp.mgrid[:crop_size, :crop_size]
        yc = yc + y0[:, None, None] * scale
        xc = xc + x0[:, None, None] * scale

        # clear invalid locations
        outputs = jnp.where(mask, outputs, 0)
        logits = jnp.where(mask, logits, 0)
        xc = jnp.where(mask, xc, -1)
        yc = jnp.where(mask, yc, -1)

        return dict(
            instance_output=outputs,
            instance_yc=yc,
            instance_xc=xc,
            instance_logit=logits,
            instance_mask=mask,
        )



import dataclasses
from functools import partial

import jax
import jax.numpy as jnp
import numpy as np
from tqdm import tqdm

_to_logits = lambda x: -jnp.log((1 / x - 1))
_EPS = jnp.finfo("float32").eps
_MIN = jnp.finfo("float32").min


@dataclasses.dataclass(frozen=True)
class HyperParams:
    n_chains: int = 4096
    n_locs: int = 512
    gamma: float = 0.1
    logit_scale: float = 4.0
    logit_offset: float = 1.0
    div_avg: float = 50.0
    div_limit: float = 0.9
    div_scale: float = 0.9
    death_avg: float = 50.0
    death_limit: float = 0
    death_scale: float = 0.9


def _get_tracking_weights(data, hp):

    yx0, yx1, logits = data
    gamma = hp.gamma

    delta = (yx1[None, :, :] - yx0[:, None, :]) ** 2
    delta = delta.sum(axis=-1)  # [n_source, n_target]

    wts = -delta / 2 * gamma * gamma
    wts = wts + logits * hp.logit_scale + hp.logit_offset

    # remove invalid
    wts = jnp.where(yx0[:, :1] >= 0, wts, _MIN)
    wts = jnp.where(yx1[:, 0] >= 0, wts, _MIN)

    return wts


def _compute_div_p(age, hp):

    return hp.div_limit / (1 + hp.div_scale ** (age - hp.div_avg))


def _compute_death_p(age, hp):

    return hp.death_limit / (1 + hp.death_scale ** (age - hp.death_avg))


@partial(jax.jit, static_argnums=3)
@partial(jax.vmap, in_axes=(0, 0, None, None))
def track_to_next_frame(key, cur_frame, data, hp):
    """
    Args
        key: rng x n_chains
        cur_frame: dict
            tracked: bool [n_chains, n_loc]
            age: int [n_chains, n_loc]
            parent: int [n_chains, n_loc]
        data: a tuple of (yx0, yx1, logits) float
        hp: parameters
    returns:
        next_frame: dict
        sample_weight: float
    """

    n_loc = hp.n_locs

    assert n_loc == cur_frame["tracked"].shape[-1]
    assert n_loc == cur_frame["age"].shape[-1]
    assert n_loc == cur_frame["parent"].shape[-1]

    key, key1, key2 = jax.random.split(key, 3)

    prev_tracked = cur_frame["tracked"]
    tracking_weights = _get_tracking_weights(data, hp)

    # death states
    death_p = _compute_death_p(cur_frame["age"], hp)
    death_p = jnp.where(prev_tracked, death_p, 1)
    to_track = jax.random.uniform(key1, [n_loc]) >= death_p

    # generate cell div states
    shuffled = jax.random.permutation(key, n_loc)
    div_p = _compute_div_p(cur_frame["age"], hp)
    div_p = jnp.where(prev_tracked, div_p, 0)
    to_div = jax.random.uniform(key2, [n_loc]) < div_p

    tracked = jnp.zeros([n_loc], dtype=bool)
    parents = jnp.zeros([n_loc], dtype=int) - 1

    def _get_next(k, tracked, parents):

        # use -inf to ensure no duplicate selections
        w = jnp.where(tracked, -jnp.inf, tracking_weights[k])
        choice = jnp.argmax(w)

        tracked = tracked.at[choice].set(True)
        parents = parents.at[choice].set(k)

        return tracked, parents

    def _scan_inner_1(k, state):

        tracked, parents = state
        k = shuffled[k]

        return jax.lax.cond(
            to_track[k],
            _get_next,
            lambda _, tracked, parents: (tracked, parents),
            k,
            tracked,
            parents,
        )

    def _scan_inner_2(k, state):

        tracked, parents = state
        k = shuffled[k]

        return jax.lax.cond(
            to_div[k],
            _get_next,
            lambda _, tracked, parents: (tracked, parents),
            k,
            tracked,
            parents,
        )

    tracked, parents = jax.lax.fori_loop(
        0,
        n_loc,
        _scan_inner_1,
        (tracked, parents),
    )
    tracked, parents = jax.lax.fori_loop(
        0,
        n_loc,
        _scan_inner_2,
        (tracked, parents),
    )
    age = jnp.where(
        to_div[parents],
        0,
        cur_frame["age"][parents] + 1,
    )
    age = jnp.where(tracked, age, -1)

    next_frame = dict(tracked=tracked, parent=parents, age=age)

    sample_weight = tracking_weights[parents, np.arange(n_loc)]
    sample_weight = sample_weight.sum(where=tracked)

    return next_frame, sample_weight


def sample_first_frame(df, n_cells, hp, frame_no=1):
    n_chains = hp.n_chains
    n_locs = hp.n_locs

    f0 = df.loc[df["frame"] == frame_no]
    scores = jax.nn.softmax(
        # _to_logits(f0["score"].to_numpy()[:n_locs]) * hp.logit_scale
        _to_logits(f0["score"].to_numpy()[:n_locs])
        * 10
    )
    scores = np.array(scores)
    yx0 = f0[["y", "x"]].to_numpy()[:n_locs]

    if len(yx0) < n_locs:
        yx0 = np.pad(yx0, [[0, n_locs - len(yx0)], [0, 0]], constant_values=-1)

    tracked = np.zeros([n_chains, n_locs], dtype=bool)
    for k in range(n_chains):
        sel = np.random.choice(
            len(scores), [n_cells], p=scores / scores.sum(), replace=False
        )
        tracked[k, sel] = True

    age = np.random.randint(0, hp.div_avg, size=[n_chains, n_locs])
    parent = np.zeros([n_chains, n_locs], dtype=int) - 1
    cur_frame = dict(
        tracked=jnp.asarray(tracked),
        parent=jnp.asarray(parent),
        age=jnp.asarray(age),
    )

    return cur_frame, yx0


def extend_chains(key, df, cur_chains, yx0, frame_no, hp):
    n_chains = hp.n_chains
    n_locs = hp.n_locs

    f1 = df.loc[df["frame"] == frame_no]
    yx1 = f1[["y", "x"]].to_numpy()[:n_locs]

    if len(yx1) < n_locs:
        yx1 = np.pad(yx1, [[0, n_locs - len(yx1)], [0, 0]], constant_values=-1)

    logits = _to_logits(f1["score"].to_numpy()[:n_locs])
    if len(logits) < n_locs:
        logits = np.pad(logits, [0, n_locs - len(logits)], constant_values=_MIN)

    cur_frame = cur_chains[-1]
    key, key2 = jax.random.split(key, 2)
    next_frame, weights = track_to_next_frame(
        jax.random.split(key2, n_chains),
        cur_frame,
        (yx0, yx1, logits),
        hp,
    )
    cur_chains.append(next_frame)

    # resample all chains

    weights = jax.nn.softmax(weights)
    rs = np.asarray(jax.random.choice(key, len(weights), [n_chains], p=weights))
    cur_chains = jax.tree_map(lambda v: v[rs], cur_chains)

    return cur_chains, yx1


def do_tracking(key, df, hp, starting_n_cell=1, first_frame=None, last_frame=None):

    if first_frame is None:
        first_frame = df["frame"].min()
    if last_frame is None:
        last_frame = df["frame"].max()

    frame, yx = sample_first_frame(df, starting_n_cell, hp, frame_no=first_frame)

    chain = [frame]
    yxs = [yx]
    for frame_no in tqdm(range(first_frame + 1, last_frame + 1)):
        chain, yx = extend_chains(
            jax.random.fold_in(key, frame_no), df, chain, yx, frame_no, hp
        )
        yxs.append(yx)

    # transpose the output and convert to np array
    chain = jax.tree_util.tree_map(lambda *v: np.stack(v), *chain)
    yxs = np.stack(yxs)

    return chain, yxs



from __future__ import annotations

import flax.linen as nn
import jax
import jax.numpy as jnp

from ..typing import *
from .common import *


class UNet(nn.Module):
    model_spec: Sequence[int] = (32, 64, 128, 256, 512)
    patch_size: int = 1
    se_ratio: int = -1

    def __post_init__(self):
        super().__post_init__()
        if not self.patch_size in [1, 2, 4]:
            raise ValueError("patch_size must be eith 1, 2 or 4")

    @nn.compact
    def __call__(self, x: ArrayLike) -> tuple[DataDict, DataDict]:

        encoder_out = []
        fs = max(self.patch_size, 3)
        st = self.patch_size

        for ch in self.model_spec:

            x = nn.Conv(ch, (fs, fs), (st, st), use_bias=False)(x)
            x = nn.GroupNorm(num_groups=ch, use_scale=False)(x)
            x = jax.nn.relu(x)

            x = nn.Conv(ch, (3, 3), use_bias=False)(x)
            x = nn.GroupNorm(num_groups=ch, use_scale=False)(x)
            x = jax.nn.relu(x)

            encoder_out.append(x)

            fs, st = 3, 2

        decoder_out = [x]

        for y in encoder_out[-2::-1]:

            ch = y.shape[-1]

            x = jax.image.resize(x, y.shape[:-1] + x.shape[-1:], "linear")
            x = jnp.concatenate([x, y], axis=-1)

            x = nn.Conv(ch, (3, 3), use_bias=False)(x)
            x = nn.GroupNorm(num_groups=ch, use_scale=False)(x)
            x = jax.nn.relu(x)

            x = nn.Conv(ch, (3, 3), use_bias=False)(x)
            x = nn.GroupNorm(num_groups=ch, use_scale=False)(x)
            x = jax.nn.relu(x)

            decoder_out.insert(0, x)

        keys = [str(lvl + self.start_level) for lvl in range(len(self.model_spec))]

        encoder_out = dict(zip(keys, encoder_out))
        decoder_out = dict(zip(keys, decoder_out))

        return encoder_out, decoder_out

    @property
    def start_level(self) -> int:
        if self.patch_size == 4:
            return 2
        elif self.patch_size == 2:
            return 1
        else:
            return 0



import dataclasses

import jax.numpy as jnp
import numpy as np
from flax.core.frozen_dict import freeze, unfreeze


def _to_str(p):
    return "".join(p.astype(int).reshape(-1).astype(str).tolist())


def format_predictions(pred, mask=None, encode_patch=True, threshold=0.5):
    """Produce more readable data from model predictions
    Args:
        pred: model output without batch dim
        mask: optional mask selecting cells
        threshold: float patch threshold
    Returns: dict(locations, scores, centroids, bboxes, encodings)
    """
    from lacss.ops import bboxes_of_patches

    patches = pred["instance_output"] >= threshold
    yc = pred["instance_yc"]
    xc = pred["instance_xc"]

    bboxes = bboxes_of_patches(pred)

    n_pixels = np.count_nonzero(patches, axis=(1, 2))
    centroids = jnp.stack(
        [
            (patches * yc).sum(axis=(1, 2)) / n_pixels,
            (patches * xc).sum(axis=(1, 2)) / n_pixels,
        ],
        axis=-1,
    )  # centroid

    outputs = dict(
        locations=pred["pred_locations"],
        scores=pred["pred_scores"],
        centroids=centroids,
        bboxes=bboxes,
    )

    is_valid = pred["instance_mask"].squeeze(axis=(-1, -2))
    is_valid &= patches.any(axis=(1, 2))  # no empty patches
    if mask is not None:
        is_valid &= mask

    outputs = {k: np.asarray(v)[is_valid] for k, v in outputs.items()}

    if encode_patch:

        encodings = []
        patches = np.asarray(patches)[is_valid]
        y0 = np.asarray(yc)[is_valid, 0, 0]
        x0 = np.asarray(xc)[is_valid, 0, 0]
        boxes = np.asarray(outputs["bboxes"]) - np.stack([y0, x0, y0, x0], axis=-1)

        for box, patch in zip(boxes, patches):
            roi = patch[box[0] : box[2], box[1] : box[3]]
            encodings.append(_to_str(roi))

        outputs["encodings"] = encodings

    return outputs


def show_images(imgs, locs=None, **kwargs):
    import matplotlib.patches
    import matplotlib.pyplot as plt

    fig, axs = plt.subplots(1, len(imgs), figsize=(4 * len(imgs), 5))
    if len(imgs) == 1:
        axs = [axs]

    for k, img in enumerate(imgs):
        axs[k].imshow(img, **kwargs)
        axs[k].axis("off")
        if locs is not None and locs[k] is not None:
            loc = np.round(locs[k]).astype(int)
            for p in loc:
                c = matplotlib.patches.Circle(
                    (p[1], p[0]), fill=False, edgecolor="white"
                )
                axs[k].add_patch(c)
    plt.tight_layout()


def dataclass_from_dict(klass, dikt):
    try:
        fieldtypes = {f.name: f.type for f in dataclasses.fields(klass)}
        return klass(**{f: dataclass_from_dict(fieldtypes[f], dikt[f]) for f in dikt})
    except:
        return dikt


def load_from_pretrained(pretrained: str):
    """Load a saved model.

    Args:
        pretrained: The url to the saved model.

    Returns: A tuple (module, parameters) representing the model.
    """
    import os

    if os.path.isdir(pretrained):
        # directory are orbax checkpoint

        from lacss.train import LacssTrainer

        trainer = LacssTrainer.from_checkpoint(pretrained)
        module = trainer.model.principal
        params = trainer.params["principal"]

    else:
        # uri or files were treated as pickled byte steam
        import pickle

        from .modules import Lacss
        from .train import Trainer

        if os.path.isfile(pretrained):
            with open(pretrained, "rb") as f:
                thingy = pickle.load(f)

        else:
            from urllib.request import Request, urlopen

            headers = {"User-Agent": "Wget/1.13.4 (linux-gnu)"}
            req = Request(url=pretrained, headers=headers)

            bytes = urlopen(req).read()
            thingy = pickle.loads(bytes)

        if isinstance(thingy, Trainer):
            module = thingy.model
            params = thingy.params

        else:
            cfg, params = thingy

            if isinstance(cfg, Lacss):
                module = cfg
            else:
                module = Lacss.from_config(cfg)

    if "params" in params and len(params) == 1:
        params = params["params"]

    return module, freeze(params)


def pack_x_y_sample_weight(x, y=None, sample_weight=None):
    """Packs user-provided data into a tuple."""
    if y is None:
        return (x,)
    elif sample_weight is None:
        return (x, y)
    else:
        return (x, y, sample_weight)


def unpack_x_y_sample_weight(data):
    """Unpacks user-provided data tuple."""
    if not isinstance(data, tuple):
        return (data, None, None)
    elif len(data) == 1:
        return (data[0], None, None)
    elif len(data) == 2:
        return (data[0], data[1], None)
    elif len(data) == 3:
        return (data[0], data[1], data[2])

    raise ValueError("Data not understood.")



from __future__ import annotations

import pathlib
import pickle
from functools import lru_cache, partial
from pathlib import Path
from typing import Any, Iterable, Iterator, Optional, Sequence, Union

import flax.linen as nn
import jax
import jax.numpy as jnp
from flax.core.frozen_dict import freeze, unfreeze
from flax.training.train_state import TrainState

from ..typing import *
from .loss import LossLog
from .strategy import JIT
from .utils import Inputs, _get_name

LOSSES = Union[LossFunc, Sequence[LossFunc]]
METRICS = Union[Metric, Sequence[Metric]]

# so that multiple calls return the same obj
# this avoids JIT when supplying partial func as args
_cached_partial = lru_cache(partial)


class Trainer:
    """A general purpose FLAX model trainer. Help avoiding most of the biolerplate code when trainning with FLAX.

    Attributes:
        model: A Flax module
        losses: A list of loss function (or other callabels).
        optimizer: An optax optimizer
        seed: RNG seed
        params: Current model parameters. This is a frozen dict. This is read/write.
        initialized: Whether the model has been initialized with an optimizer and initial weights.

    Example:

        ```
        trainer = lacss.train.Trainer(my_module, my_loss_func)

        trainer.initialize(my_dataset, tx=my_optimzier)

        train_it = trainer.train(my_dataset)

        for k in range(train_steps):
            loss_logs = next(train_it)
            if k % 1000 == 0:
                print(loss_logs)
                trainer.reset()
        ```

    """

    def __init__(
        self,
        model: nn.Module,
        losses: Optional[LOSSES] = None,
        optimizer: Optional[Optimizer] = None,
        seed: Union[int, Array] = 42,
        strategy: type = JIT,
    ):
        """Constructor

        Args:
            model: A Flax module. It can be a bound module with parameters.
            losses: A loss function (callabels) or a list of loss functions, or None.
                These should have the call signiture of:
                    loss = loss_fn(inputs, labels, preds)
                The "preds" is the model output. The "inputs" and "labels" are from the train dataset.
                The model trains on the sum of all losses.
            optimizer: An optax optimzier
            seed: RNG seed.
            strategy: Training backend. See [Traing backends](./#training-backends).

        """
        self.model = model
        self.losses = losses

        self.seed = seed if isinstance(seed, jnp.ndarray) else jax.random.PRNGKey(seed)
        self._optimizer = optimizer

        self._strategy = strategy
        self._initialized = False

    def reset(
        self,
        loss_weights: Optional[Sequence[float]] = None,
        losses: Optional[LOSSES] = None,
    ):
        """Reset internal loss value tracking

        Args:
            loss_weights: Optional weights of individual loss functions. If not None, the
                total loss is weighted sum.
            losses: Optionally override the class losses defined as the object initialization.
        """
        if losses is None:
            losses = self.losses

        if losses is None:
            raise ValueError(f"No loss functions provided")

        try:
            iter(losses)
        except:
            losses = (losses,)

        if loss_weights is None:
            loss_weights = (1.0,) * len(losses)

        if len(loss_weights) != len(losses):
            raise ValueError(
                f"Loss weights supplied {loss_weights} does not match the number of loss functions ({len(losses)})"
            )

        self.loss_logs = tuple(
            LossLog(loss, w) for loss, w in zip(losses, loss_weights)
        )

    @property
    def initialized(self):
        return self._initialized

    def initialize(self, data: Iterable, tx: Optional[Optimizer] = None) -> None:
        """Initialize the model weights and optimizer states.

        Args:
            data: An iterator or iterable to produce training dataset. It is not
                used if model is bound with weights already.
                see [train()](./#lacss.train.base_trainer.Trainer.train)
            tx: Optional optax optimzier for when the object was constructed without an
                optimizer
        """
        if tx is None:
            tx = self.optimizer

        if not self._initialized:
            if self.model.scope is None:
                peek = next(iter(data))
                # inputs, _, _ = unpack_x_y_sample_weight(peek)
                inputs = peek[0] if isinstance(peek, tuple) else peek

                self.seed, key = jax.random.split(self.seed)
                variables = self._strategy.init_fn(key, self.model, inputs)

            else:
                self.model, variables = self.model.unbind()

            self.state = TrainState.create(
                apply_fn=self.model.apply,
                params=variables["params"],
                tx=tx,
            )

        else:
            raise ValueError("Calling initialize() on already initialized Trainer")

        self._initialized = True

    def __call__(
        self, inputs: Any, *, strategy: Optional[type] = None, **kwargs
    ) -> Any:
        """Calling the underlying model

        Args:
            inputs: A tuple or a dict as the inputs for the model.
            strategy: Optionally override the default strategy.
            **kwargs: Additional keyword args passed on to the model

        Returns:
            Model outputs.
        """
        if strategy is None:
            strategy = self._strategy

        predict_fn = strategy.predict

        state = self.state.replace(apply_fn=_cached_partial(self.model.apply, **kwargs))
        preds = predict_fn(state, inputs)

        return preds

    def compute_loss_log(self) -> dict:
        return {
            _get_name(loss_log.loss_fn): loss_log.compute()
            for loss_log in self.loss_logs
        }

    def train(
        self,
        dataset: Iterable,
        strategy: Optional[type] = None,
        rng_cols: Sequence[str] = None,
        **kwargs,
    ) -> Iterator:
        """Create the training iterator

        Args:
            dataset: An iterator or iterable to supply the training data.
                The dataset should produce ```(inputs, labels, sample_weight)```, however
                both the labels and the sample_weight are optional. The inputs is either a tuple
                or a dict. If the inputs is a dict, the keys are interpreted as the names for
                keyword args of the model's __call__ function.
            strategy: Optionally override the default strategy.
            rng_cols: Names of any RNG used by the model. Should be a list of strings.
            **kwargs: Additional keyward args passed to the model. E.g. "training=True"

        Returns:
            A iterator. Stepping through the iterator will train the model. The iterator
                itself returns a loss log dict, which are mean loss values for each loss
                function.

        """
        if strategy is None:
            strategy = self._strategy

        if not self._initialized:
            raise ValueError("Try to run uninitialized trainer")

        train_fn = strategy.train_step

        self.state = self.state.replace(
            apply_fn=_cached_partial(self.model.apply, **kwargs)
        )

        self.reset()

        self.seed, seed = jax.random.split(self.seed)
        for step, data in enumerate(dataset):
            # batch = unpack_x_y_sample_weight(data)
            batch = data

            if rng_cols is not None:
                key = jax.random.fold_in(seed, step)
                keys = jax.random.split(key, len(rng_cols))
                rngs = {name: k for name, k in zip(rng_cols, keys)}
            else:
                rngs = None

            self.state, self.loss_logs, preds = train_fn(
                self.state, self.loss_logs, batch, rngs
            )

            batch_logs = self.compute_loss_log()

            yield batch_logs

    def save_model(self, path: PathLike, sub_module: Optional[str] = None) -> None:
        """Save the model in a pickled file. The pickle is a tuple of
            (module, weights).

        Args:
            path: The file path.
            sub_module: Optionally only save a sub_module of the model
                by specifying the name
        """
        module = self.model
        params = self.params

        if sub_module is not None:
            module = module.bind(dict(params=params))
            module = getattr(module, sub_module)
            module, params = module.unbind()
            params = params["params"]

        if isinstance(path, str):
            path = Path(path)

        path.parent.mkdir(parents=True, exist_ok=True)

        with open(path, "wb") as f:
            pickle.dump((module, params), f)

    def pickle_self(self, path: PathLike) -> None:
        """Make a pickle save of the trainer. This saves the model as well as
        the training states.

        Args:
            path: The file path.
        """
        path = pathlib.Path(path)
        path.parent.mkdir(parents=True, exist_ok=True)

        with open(path, "wb") as f:
            pickle.dump(self, f)

    @staticmethod
    def restore_from_pickle(path: PathLike):
        """Restore from the a pickled saved.

        Args:
            path: The pickle file path.

        Returns:
            A new trainer object.
        """
        path = pathlib.Path(path)

        try:
            _bytes = path.read_bytes()
        except BaseException as e:
            raise OSError(f"Could not load the checkpoint. Got exception: {e}")

        trainer = pickle.loads(_bytes)

        if not isinstance(trainer, Trainer):
            raise TypeError("The saved obj is not a Trainer checkpoint")

        return trainer

    def test(
        self,
        dataset: Iterable,
        metrics: METRICS,
        strategy: Optional[type] = None,
        **kwargs,
    ) -> Iterator:
        """Create test/validation iterator.

        Args:
            dataset: An iterator or iterable to supply the testing data.
                The iterator should yield a tupple of (inputs, labels).
            metrics: A list of Metric objects. They should have two functions:
                m.update(preds, **kwargs):
                    preds is the model output. the remaining kwargs are content of
                    labels.
                m.compute():
                    which should return the accumulated metric value.
            strategy: Optionally override the default strategy.

        Returns:
            An iterator. Stepping through it will drive the updating of each metric
                obj. The iterator itself return the list of metrics.
        """
        if strategy is None:
            strategy = self._strategy

        try:
            iter(metrics)
        except TypeError:
            metrics = [metrics]

        state = self.state.replace(apply_fn=_cached_partial(self.model.apply, **kwargs))
        predict_fn = strategy.predict

        for data in dataset:
            # inputs, labels, _ = unpack_x_y_sample_weight(data)
            inputs = data[0] if isinstance(data, tuple) else data
            preds = predict_fn(state, inputs)
            kwargs = dict(
                batch=data,
                prediction=preds,
            )
            for m in metrics:
                m.update(**kwargs)

            yield metrics

    def test_and_compute(self, *args, **kwargs) -> dict:
        """A convient function to compute all metrics. See [test() fucntion](./#lacss.train.base_trainer.Trainer.test)

        Returns:
            A metric dict. Keys are metric names.
        """
        for metrics in self.test(*args, **kwargs):
            pass
        return {_get_name(m): m.compute() for m in metrics}

    @property
    def params(self) -> Params:
        return self.state.params

    @params.setter
    def params(self, new_params: Params) -> None:
        old_state = self.state
        self.state = TrainState.create(
            apply_fn=self.model.apply,
            params=new_params,
            tx=old_state.tx,
        )

    @property
    def optimizer(self) -> Optimizer:
        return self._optimizer

    @optimizer.setter
    def optimizer(self, tx: Optimizer) -> None:
        self._optimizer = tx

        if self._initialized:
            self.state = TrainState.create(
                apply_fn=self.model.apply,
                params=self.params,
                tx=tx,
            )



from __future__ import annotations

from functools import partial

import jax.numpy as jnp

from ..typing import *


def generalized_binary_loss(
    pred: ArrayLike, gt: ArrayLike, gamma: int | float, beta: int | float
) -> Array:
    EPS = jnp.finfo(jnp.promote_types(pred, gt)).eps

    p_t = pred * gt + (1 - pred) * (1 - gt)

    ff = 1 - p_t
    ce = -jnp.log(jnp.clip(p_t, EPS, 1))

    loss = (ff**gamma) * (ce**beta)

    return loss


binary_focal_factor_loss = partial(generalized_binary_loss, beta=0)

binary_focal_crossentropy = partial(generalized_binary_loss, beta=1)

binary_crossentropy = partial(generalized_binary_loss, gamma=0, beta=1)


def mean_over_boolean_mask(loss: ArrayLike, mask: ArrayLike) -> Array:
    mask = mask.reshape(mask.shape[0], 1)
    n_instances = jnp.count_nonzero(mask) + 1e-8

    loss = loss.reshape(loss.shape[0], -1)
    loss = loss.mean(axis=1, keepdims=True).sum(where=mask)
    loss /= n_instances

    return loss



# -*- coding: utf-8 -*-
# Slightly modified from https://github.com/tmoopenn/seq-nms

import numpy as np
import copy 

from lacss.ops import box_intersection

# from compute_overlap import compute_overlap_areas_given, compute_area
def compute_area(boxes):
    return (boxes[:, 2] - boxes[:,0]) * (boxes[:, 3] - boxes[:,1])

def compute_overlap_areas_given(boxes1, boxes2, area2):
    its = box_intersection(boxes1, boxes2)
    area1 = compute_area(boxes1)
    # area2 = compute_area(boxes2)
    iou = its / (area1[:, None] + area2 - its + 1e-8)
    return iou

def seq_nms(boxes, scores, labels=[], linkage_threshold=0.6, nms_threshold=0.4, score_metric='avg'):
    ''' Filter detections using the seq-nms algorithm. Boxes and classifications should be organized sequentially along the first dimension 
    corresponding to the input frame.  
    Args 
        boxes                 : Tensor of shape (num_frames, num_boxes, 4) containing the boxes in (x1, y1, x2, y2) format.
        scores                : Tensor of shape (num_frames, num_boxes) containing the confidence score for each box.
        linkage_threshold     : Threshold used to link two boxes in adjacent frames 
        nms_threshold         : Threshold for the IoU value to determine when a box should be suppressed with regards to a best sequence.
    '''
    # use filtered boxes and scores to create nms graph across frames 
    box_graph = build_box_sequences(boxes, scores, labels, linkage_threshold=linkage_threshold)
    print("BOX GRAPH SHAPE", box_graph.shape)
    _seq_nms(box_graph, boxes, scores, nms_threshold, score_metric=score_metric)

def build_box_sequences(boxes, scores, labels=[], linkage_threshold=0.5):
    ''' Build bounding box sequences across frames. A sequence is a set of boxes that are linked in a video
    where we define a linkage as boxes in adjacent frames (of the same class) with IoU above linkage_threshold (0.5 by default).
    Args
        boxes                  : Tensor of shape (num_frames, num_boxes, 4) containing the boxes in (x1, y1, x2, y2) format. 
        scores                : Tensor of shape (num_frames, num_boxes) containing the confidence score for each box.
        linkage_threshold      : Threshold for the IoU value to determine if two boxes in neighboring frames are linked 
    Returns 
        A list of shape (num_frames - 1, num_boxes, k, 1) where k is the number of edges to boxes in neighboring frame (s.t. 0 <= k <= num_boxes at f+1)
        and last dimension gives the index of that neighboring box. 
    '''
    box_graph = []
    # iterate over neighboring frames 
    for f in range(boxes.shape[0] - 1):
        boxes_f, scores_f = boxes[f,:,:], scores[f,:]
        boxes_f1, scores_f1 = boxes[f+1,:,:], scores[f+1,:]
        if f == 0:
            areas_f = compute_area(boxes_f.astype(np.double)) #(boxes_f[:,2] - boxes_f[:,0] + 1) * (boxes_f[:,3] - boxes_f[:,1] + 1)
        else: 
            areas_f = areas_f1

        # calculate areas for boxes in next frame
        areas_f1 = compute_area(boxes_f1.astype(np.double)) #(boxes_f1[:,2] - boxes_f1[:,0] + 1) * (boxes_f1[:,3] - boxes_f1[:,1] + 1)

        adjacency_matrix = []
        for i, box in enumerate(boxes_f):
            overlaps = compute_overlap_areas_given(np.expand_dims(box,axis=0).astype(np.double), boxes_f1.astype(np.double), areas_f1.astype(np.double) )[0]

            # add linkage if IoU greater than threshold and boxes have same labels i.e class  
            if len(labels) == 0 :
                edges = [ovr_idx for ovr_idx, IoU in enumerate(overlaps) if IoU >= linkage_threshold]
            else:
                edges = [ovr_idx for ovr_idx, IoU in enumerate(overlaps) if IoU >= linkage_threshold and labels[f,i] == labels[f+1,ovr_idx]]
            adjacency_matrix.append(edges)
        box_graph.append(adjacency_matrix)
    return np.array(box_graph, dtype="object")

def find_best_sequence(box_graph, scores):
    ''' Given graph of all linked boxes, find the best sequence in the graph. The best sequence 
    is defined as the sequence with the maximum score across an arbitrary number of frames.
    We build the sequences back to front from the last frame to easily capture start of new sequences/
    Condition to start of new sequence: 
        if there are no edges from boxes in frame t-1, then the box in frame t must be the start of a new sequence.
        This assumption is valid since all scores are positive so we can always improve a sequence by increasing its length. 
        Therefore if there are links to a box from previous frames, we can always build a better path by extending it s.t. 
        the box cannot be the start of a new best sequence. 
    Args
        box_graph             : list of shape (num_frames - 1, num_boxes, k) returned from build_box_sequences that contains box sequences 
        scores                : Tensor of shape (num_frames, num_boxes) containing the label for the corresponding box. 
    Returns 
        None
    '''
    # list of tuples storing (score up to current frame, path up to current frame)
    # we dynamically build up best paths through graph starting from the end frame
    # s.t we can determine the beginning of sequences i.e. if there are no links 
    # to a box from previous frames, then it is a candidate for starting a sequence 
    max_scores_paths = [] 

    # list of all independent sequences where a given row corresponds to starting frame
    sequence_roots = []

    # starting from the last frame, build base paths i.e paths consisting of a single node 
    max_scores_paths.append([(score, [idx]) for idx, score in enumerate(scores[-1])])

    for reverse_idx, frame_edges in enumerate(box_graph[::-1]): # list of edges between neigboring frames i.e frame dimension 
        max_paths_f = []
        used_in_sequence = np.zeros(len(max_scores_paths[-1]), int)
        frame_idx = len(box_graph) - reverse_idx - 1
        for box_idx, box_edges in enumerate(frame_edges): # list of edges for each box in frame i.e. box dimension
            if not box_edges: # no edges for current box so consider it a max path consisting of a single node 
                max_paths_f.append((scores[frame_idx][box_idx], [box_idx]))
            else: # extend previous max paths 
                # here we use box_edges list to index used_in_sequence list and mark boxes in corresponding frame t+1 
                # as part of a sequence since we have links to them and can always make a better max path by making it longer (no negative scores)
                used_in_sequence[box_edges] = 1
                prev_idx = np.argmax([max_scores_paths[-1][bidx][0] for bidx in box_edges])
                score_so_far = max_scores_paths[-1][box_edges[prev_idx]][0]
                path_so_far = copy.copy(max_scores_paths[-1][box_edges[prev_idx]][1])
                path_so_far.append(box_idx)
                max_paths_f.append((scores[frame_idx][box_idx] + score_so_far, path_so_far))
        
        # create new sequence roots for boxes in frame at frame_idx + 1 that did not have links from boxes in frame_idx
        new_sequence_roots = [max_scores_paths[-1][idx] for idx, flag in enumerate(used_in_sequence) if flag == 0]

        sequence_roots.append(new_sequence_roots) 
        max_scores_paths.append(max_paths_f)
    
    # add sequences starting in begining frame as roots 
    sequence_roots.append(max_scores_paths[-1])

    # reverse sequence roots since built sequences from back to front 
    sequence_roots = sequence_roots[::-1]

    # iterate sequence roots to find sequence with max score 
    best_score = 0 
    best_sequence = [] 
    sequence_frame_index = 0
    for index, frame_sequences in enumerate(sequence_roots):
        if not frame_sequences: continue 
        max_index = np.argmax([sequence[0] for sequence in frame_sequences])
        if frame_sequences[max_index][0] > best_score:
            best_score = frame_sequences[max_index][0]
            best_sequence = frame_sequences[max_index][1][::-1] # reverse path 
            sequence_frame_index = index
    return sequence_frame_index, best_sequence, best_score


def rescore_sequence(sequence, scores, sequence_frame_index, max_sum, score_metric='avg'):
    ''' Given a sequence, rescore the confidence scores according to the score_metric.
    Args
        sequence                    : The best sequence containing indices of boxes 
        scores                      : Tensor of shape (num_frames, num_boxes) containing the label for the corresponding box. 
        sequence_frame_index        : The index of the frame where the best sequence begins 
        best_score                  : The summed score of boxes in the sequence 
    Returns 
        None   
    '''
    if score_metric == 'avg':
        avg_score=max_sum/len(sequence)
        for i,box_ind in enumerate(sequence):
            scores[sequence_frame_index+i][box_ind]= avg_score
    elif score_metric == 'max':
        max_score = 0.0
        for i, box_ind in enumerate(sequence):
            if scores[sequence_frame_index + i][box_ind] > max_score: max_score = scores[sequence_frame_index+i][box_ind]
        for i, box_ind in enumerate(sequence):
            scores[sequence_frame_index + i][box_ind] = max_score
    else:
        raise ValueError("Invalid score metric")

def delete_sequence(sequence_to_delete, sequence_frame_index, scores, boxes, box_graph, suppress_threshold=0.3):
    ''' Given a sequence, remove its connections in box graph (create graph of linked boxes across frames).
    Args
        sequence_to_delete          : The best sequence containing indices of boxes to be deleted
        sequence_frame_index        : The index of the frame where the best sequence begins 
        scores                      : Tensor of shape (num_frames, num_boxes) containing the label for the corresponding box. 
        boxes                       : Tensor of shape (num_frames, num_boxes, 4) containing the boxes in (x1, y1, x2, y2) format.
        box_graph                   : list of shape (num_frames - 1, num_boxes, k) returned from build_box_sequences that contains box sequences 
        suppress_threshold          : Threshold for suprresing boxes that have an IoU with sequence boxes greater than the threshold 
    Returns 
        None  
    '''
    for i,box_idx in enumerate(sequence_to_delete):
        other_boxes = boxes[sequence_frame_index + i]
        box_areas = compute_area(other_boxes.astype(np.double))
        seq_box_area = box_areas[box_idx]
        seq_box = boxes[sequence_frame_index+i][box_idx]

        overlaps = compute_overlap_areas_given(np.expand_dims(seq_box,axis=0).astype(np.double), boxes[sequence_frame_index+i,:].astype(np.double), box_areas.astype(np.double))[0]
        deletes=[ovr_idx for ovr_idx,IoU in enumerate(overlaps) if IoU >= suppress_threshold]

        if sequence_frame_index + i < len(box_graph): 
            
            for delete_idx in deletes:
                box_graph[sequence_frame_index+i][delete_idx]=[]

        if i > 0 or sequence_frame_index > 0:
            # remove connections to current sequence node from previous frame nodes
            for prior_box in box_graph[sequence_frame_index+i-1]: 
                for delete_idx in deletes:
                    if delete_idx in prior_box:
                        prior_box.remove(delete_idx)
    
        deletes.remove(box_idx)
        scores[sequence_frame_index+i, deletes] = -1


def _seq_nms(box_graph, boxes, scores, nms_threshold, score_metric='avg'):
    ''' Iteratively executes the seq-nms algorithm given a box graph.
    Args
        box_graph                   : list of shape (num_frames - 1, num_boxes, k) returned from build_box_sequences that contains box sequences 
        boxes                       : Tensor of shape (num_frames, num_boxes, 4) containing the boxes in (x1, y1, x2, y2) format.
        scores                      : Tensor of shape (num_frames, num_boxes) containing the label for the corresponding box. 
        nms_threshold               : Threshold for the IoU value to determine when a box should be suppressed with regards to a best sequence.
    Returns 
        None
    '''
    while True: 
        sequence_frame_index, best_sequence, best_score = find_best_sequence(box_graph, scores)
        # print(best_sequence, best_score)
        if len(best_sequence) <= 1:
            break 
        rescore_sequence(best_sequence, scores, sequence_frame_index, best_score, score_metric=score_metric)
        delete_sequence(best_sequence, sequence_frame_index, scores, boxes, box_graph, suppress_threshold=nms_threshold)



from __future__ import annotations

from typing import Optional, Sequence

import flax.linen as nn
import jax
import jax.numpy as jnp

from ..typing import DataDict, Params
from .common import *

""" Implements the convnext encoder. Described in https://arxiv.org/abs/2201.03545
Original implementation: https://github.com/facebookresearch/ConvNeXt
"""


class _Block(nn.Module):
    """ConvNeXt Block.
    Args:
        drop_path (float): Stochastic depth rate. Default: 0.0
        layer_scale_init_value (float): Init value for Layer Scale. Default: 1e-6.
    """

    drop_rate: int = 0.4
    layer_scale_init_value: float = 1e-6
    kernel_size: int = 7

    @nn.compact
    def __call__(self, x: ArrayLike, *, training: Optional[bool] = None) -> Array:
        dim = x.shape[-1]
        ks = self.kernel_size
        scale = self.layer_scale_init_value

        shortcut = x

        x = nn.Conv(dim, (ks, ks), feature_group_count=dim)(x)
        x = nn.LayerNorm(epsilon=1e-6)(x)
        x = nn.Dense(dim * 4)(x)
        x = jax.nn.gelu(x)
        x = nn.Dense(dim)(x)

        if scale > 0:
            gamma = self.param(
                "gamma", lambda rng, shape: scale * jnp.ones(shape), (x.shape[-1])
            )
            x = x * gamma

        deterministic = training is None or not training
        x = DropPath(self.drop_rate)(x, deterministic=deterministic)

        x = x + shortcut

        return x


# utility funcs

_imagenet_weights_urls = {
    "convnext_tiny_1k": "https://dl.fbaipublicfiles.com/convnext/convnext_tiny_1k_224_ema.pth",
    "convnext_small_1k": "https://dl.fbaipublicfiles.com/convnext/convnext_small_1k_224_ema.pth",
    "convnext_base_1k": "https://dl.fbaipublicfiles.com/convnext/convnext_base_1k_224_ema.pth",
    "convnext_large_1k": "https://dl.fbaipublicfiles.com/convnext/convnext_large_1k_224_ema.pth",
    "convnext_tiny_22k": "https://dl.fbaipublicfiles.com/convnext/convnext_tiny_22k_224.pth",
    "convnext_small_22k": "https://dl.fbaipublicfiles.com/convnext/convnext_small_22k_224.pth",
    "convnext_base_22k": "https://dl.fbaipublicfiles.com/convnext/convnext_base_22k_224.pth",
    "convnext_large_22k": "https://dl.fbaipublicfiles.com/convnext/convnext_large_22k_224.pth",
    "convnext_xlarge_22k": "https://dl.fbaipublicfiles.com/convnext/convnext_xlarge_22k_224.pth",
}


class ConvNeXt(nn.Module):
    """ConvNeXt CNN backbone

    Attributes:
        patch_size: Stem patch size
        depths: Number of blocks at each stage.
        dims: Feature dimension at each stage.
        drop_path_rate: Stochastic depth rate.
        layer_scale_init_value: Init value for Layer Scale.
        out_channels:
            FPN output channels. Setting this to -1 disable the FPN, in which case
            the model output only encoder outputs.
    """

    patch_size: int = 4
    depths: Sequence[int] = (3, 3, 27, 3)
    dims: Sequence[int] = (96, 192, 384, 768)
    drop_path_rate: float = 0.0
    layer_scale_init_value: float = 1e-6
    out_channels: int = 384

    @nn.compact
    def __call__(
        self, x: ArrayLike, *, training: Optional[bool] = None
    ) -> tuple[DataDict, DataDict]:
        """
        Args:
            x: Image input.
            training: Whether run the network in training mode (i.e. with stochastic depth)

        Returns:
            A tuple of (encoder_outputs, decoder_outputs). Both are dictionaries mapping
                feature scale (e.g. "2") to features. If out_channels is -1, the decoder_output
                is None.
        """
        dp_rate = 0
        outputs = []
        for k in range(len(self.depths)):
            if k == 0:
                ps = self.patch_size
                x = nn.Conv(self.dims[k], (ps, ps), strides=(ps, ps))(x)
                x = nn.LayerNorm(epsilon=1e-6)(x)
            else:
                x = nn.LayerNorm(epsilon=1e-6)(x)
                x = nn.Conv(self.dims[k], (2, 2), strides=(2, 2))(x)

            for _ in range(self.depths[k]):
                x = _Block(dp_rate, self.layer_scale_init_value)(x, training=training)
                dp_rate += self.drop_path_rate / (sum(self.depths) - 1)

            outputs.append(x)

        keys = [str(k + 1 if self.patch_size == 2 else k + 2) for k in range(4)]
        encoder_out = dict(zip(keys, outputs))

        if self.out_channels > 0:
            decoder_out = FPN(self.out_channels)(outputs)
            decoder_out = dict(zip(keys, decoder_out))
        else:
            decoder_out = None

        return encoder_out, decoder_out

    def _load_weight(self, jax_params, url):
        import torch
        from flax.core.frozen_dict import freeze, unfreeze

        def t(m, new_value):
            new_value = jnp.array(new_value)
            assert m.shape == new_value.shape
            m = new_value

        checkpoint = torch.hub.load_state_dict_from_url(url=url, map_location="cpu")
        params = checkpoint["model"]
        jax_params = unfreeze(jax_params)

        cnt = 0
        for i in range(4):
            for k in range(self.depths[i]):
                block = jax_params[f"_Block_{cnt}"]
                t(block["gamma"], params[f"stages.{i}.{k}.gamma"])
                t(block["Conv_0"]["bias"], params[f"stages.{i}.{k}.dwconv.bias"])
                t(
                    block["Conv_0"]["kernel"],
                    params[f"stages.{i}.{k}.dwconv.weight"].permute(2, 3, 1, 0),
                )
                t(block["LayerNorm_0"]["bias"], params[f"stages.{i}.{k}.norm.bias"])
                t(block["LayerNorm_0"]["scale"], params[f"stages.{i}.{k}.norm.weight"])
                t(block["Dense_0"]["bias"], params[f"stages.{i}.{k}.pwconv1.bias"])
                t(
                    block["Dense_0"]["kernel"],
                    params[f"stages.{i}.{k}.pwconv1.weight"].transpose(0, 1),
                )
                t(block["Dense_1"]["bias"], params[f"stages.{i}.{k}.pwconv2.bias"])
                t(
                    block["Dense_1"]["kernel"],
                    params[f"stages.{i}.{k}.pwconv2.weight"].transpose(0, 1),
                )

                cnt += 1

        norm = jax_params[f"LayerNorm_0"]
        t(norm["bias"], params["downsample_layers.0.1.bias"])
        t(norm["scale"], params["downsample_layers.0.1.weight"])

        conv = jax_params[f"Conv_0"]
        t(conv["bias"], params["downsample_layers.0.0.bias"])
        t(conv["kernel"], params["downsample_layers.0.0.weight"].permute(2, 3, 1, 0))

        for i in range(1, 4):
            norm = jax_params[f"LayerNorm_{i}"]
            t(norm["bias"], params[f"downsample_layers.{i}.0.bias"])
            t(norm["scale"], params[f"downsample_layers.{i}.0.weight"])
            conv = jax_params[f"Conv_{i}"]
            t(conv["bias"], params[f"downsample_layers.{i}.1.bias"])
            t(
                conv["kernel"],
                params[f"downsample_layers.{i}.1.weight"].permute(2, 3, 1, 0),
            )

        return freeze(jax_params)

    def get_imagenet_weights(self, model_type: str) -> Params:
        """Get imagenet weights

        Args:
            model_type: The expected model specification. This must match the current
                instance attributes.

                * tiny: dims=(96, 192, 384, 768), depths=(3,3,9,3)
                * small: dims=(96, 192, 384, 768), depths=(3,3,27,3)
                * base: dims=(128, 256, 512, 1024), depths=(3,3,27,3)
                * large: dims=(192, 384, 768, 1536), depths=(3,3,27,3)
                * X-large: dims=(256, 512, 1024, 2048), depths=(3,3,27,3)

        Returns:
            A frozen dict representing weights of the current module
        """

        init_params = self.init(jax.random.PRNGKey(0), jnp.zeros([64, 64, 3]))
        init_params = init_params["params"]
        params = self._load_weight(
            init_params,
            _imagenet_weights_urls[f"convnext_{model_type}_22k"],
        )

        return params



""" 
Attributes:
    model_urls: URLs for build-in pretrain models. e.g model_urls["livecell"].
"""
from __future__ import annotations

import logging
import pickle
from functools import lru_cache, partial
from typing import Mapping, Optional, Sequence, Tuple, Union

import flax.linen as nn
import jax
import jax.numpy as jnp
import numpy as np
from flax.core.frozen_dict import freeze, unfreeze

import lacss.modules
import lacss.ops

from ..ops import patches_to_label, sorted_non_max_suppression
from ..typing import *
from ..utils import load_from_pretrained

Shape = Sequence[int]

_cached_partial = lru_cache(partial)

model_urls: Mapping[str, str] = {
    "livecell": "https://data.mendeley.com/public-files/datasets/sj3vrvm6w3/files/439e524f-e4e9-4f97-9f38-c22cb85adbd1/file_downloaded",
    "tissuenet": "https://data.mendeley.com/public-files/datasets/sj3vrvm6w3/files/1e0a839d-f564-4ee0-a4f3-34792df7c613/file_downloaded",
}


def _remove_edge_instances(
    pred: dict,
    img_shape: Tuple[int, int],
    remove_top: bool = True,
    remove_bottom: bool = True,
    remove_left: bool = True,
    remove_right: bool = True,
    threshold: float = 0.5,
    min_border_pixel: int = 1,
):
    h, w = img_shape
    segs = pred["instance_output"] >= threshold
    yc = pred["instance_yc"]
    xc = pred["instance_xc"]

    removal = jnp.zeros([len(segs)], dtype=bool)
    if remove_top:
        is_top = jnp.count_nonzero((yc <= 0) & segs, axis=(1, 2)) >= min_border_pixel
        removal |= is_top
    if remove_bottom:
        is_bottom = (
            jnp.count_nonzero((yc >= h - 1) & segs, axis=(1, 2)) >= min_border_pixel
        )
        removal |= is_bottom
    if remove_left:
        is_left = jnp.count_nonzero((xc <= 0) & segs, axis=(1, 2)) >= min_border_pixel
        removal |= is_left
    if remove_right:
        is_right = (
            jnp.count_nonzero((xc >= w - 1) & segs, axis=(1, 2)) >= min_border_pixel
        )
        removal |= is_right

    removal = removal.reshape(pred["instance_mask"].shape)
    pred["instance_mask"] &= ~removal

    return pred


@partial(jax.jit, static_argnums=(0, 3))
def _predict(apply_fn, params, image, scaling):
    if scaling != 1:
        h, w, c = image.shape
        image = jax.image.resize(
            image, [round(h * scaling), round(w * scaling), c], "linear"
        )

    preds = apply_fn(dict(params=params), image)
    preds["pred_bboxes"] = lacss.ops.bboxes_of_patches(preds)

    del preds["encoder_features"]
    del preds["decoder_features"]
    del preds["lpn_features"]
    del preds["lpn_scores"]
    del preds["lpn_regressions"]
    del preds["instance_logit"]

    if scaling != 1:
        (
            preds["instance_output"],
            preds["instance_yc"],
            preds["instance_xc"],
        ) = lacss.ops.rescale_patches(preds, 1 / scaling)

        preds["pred_locations"] = preds["pred_locations"] / scaling
        preds["pred_bboxes"] = preds["pred_bboxes"] / scaling

    return preds


class Predictor:
    """Main class interface for model deployment. This is the only class you
    need if you don't train your own model

    Examples:
        The most common use case is to use a build-in pretrained model.

            import lacss.deploy

            # look up the url of a build-in mode
            url = lacss.deploy.model_urls["livecell"]

            # create the predictor instance
            predictor = lacss.deploy.Predictor(url)

            # make a prediction
            label = predictor.predict_label(image)

    Attributes:
        module: The underlying FLAX module
        params: Model weights.
        detector: The detector submodule for convinence. A common use case
            is to customize this submodule during inference. e.g.
            ```
            predictor.detector.test_min_score = 0.5
            ```
            Detector submodule has no trained paramters.
    """

    def __init__(self, url: str, precompile_shape: Optional[Shape] = None):
        """Construct Predictor

        Args:
            url: A URL or local path to the saved model.
                URLs for build-in pretrained models can be found in lacss.deploy.model_urls
            precompile_shape: Image shape(s) for precompiling. Otherwise
                the model will be recompiled for every new input image shape.

        """

        if isinstance(url, tuple) and len(url) == 2:
            self.model = url
            if not isinstance(self.model[0], nn.Module):
                raise ValueError(
                    "Initiaize the Predictor with a tuple, but the first element is not a Module."
                )
        else:
            self.model = load_from_pretrained(url)

        if precompile_shape is not None:
            logging.info("Prcompile the predictor for image shape {precompile_shape}.")
            logging.info("This will take several minutes.")

            try:
                iter(precompile_shape[0])
            except:
                precompile_shape = [precompile_shape]
            for shape in precompile_shape:
                x = jnp.zeros(shape)
                _ = self.predict(x)

    def predict(
        self,
        image: ArrayLike,
        *,
        min_area: float = 0,
        remove_out_of_bound: bool = False,
        scaling: float = 1.0,
        return_label: bool = False,
        **kwargs,
    ) -> Union[dict, Array]:
        """Predict segmentation.

        Args:
            image: A ndarray of (h,w,c) format. Value of c must be 1-3

        Keyword Args:
            min_area: Minimum area of a valid prediction.
            remove_out_of_bound: Whether to remove out-of-bound predictions. Default is False.
            scaling: A image scaling factor. If not 1, the input image will be resized internally before fed
                to the model. The results will be resized back to the scale of the orginal input image.
            return_label: Whether to output full model prediction or segmentation label.

        Returns:

            If return_label is False, returns model predictions with following elements:

                - pred_scores: The prediction scores of each instance.
                - pred_bboxes: The bounding-boxes of detected instances in y0x0y1x1 format
                - instance_output: A 3d array representing segmentation instances.
                - instance_yc: The meshgrid y-coordinates of the instances.
                - instance_xc: The meshgrid x-coordinates of the instances.
                - instance_mask: a masking tensor. Invalid instances are maked with 0.

            otherwise return a segmentation label.
        """

        module, params = self.model

        if len(kwargs) > 0:
            apply_fn = _cached_partial(module.apply, **kwargs)
        else:
            apply_fn = module.apply

        preds = _predict(apply_fn, params, image, scaling)

        # check prediction validity
        instance_mask = preds["instance_mask"].squeeze(axis=(1, 2))

        if min_area > 0:
            areas = np.count_nonzero(preds["instance_output"] > 0.5, axis=(1, 2))
            instance_mask &= areas > min_area

        if remove_out_of_bound:
            pred_locations = preds["pred_locations"]
            h, w, _ = image.shape
            valid_locs = (pred_locations >= 0).all(axis=-1)
            valid_locs &= pred_locations[:, 0] < h
            valid_locs &= pred_locations[:, 1] < w
            instance_mask &= valid_locs

        preds["instance_mask"] = instance_mask.reshape(-1, 1, 1)

        if return_label:
            return patches_to_label(
                preds,
                input_size=image.shape[:2],
                score_threshold=0.5,
            )
        else:
            return preds

    def predict_label(
        self, image: ArrayLike, *, score_threshold: float = 0.5, **kwargs
    ) -> Array:
        """Predict segmentation in image label format

        Args:
            image: Input image.
            score_threshold: The minimal prediction scores.

        Returns:
            A [H, W] array. The values indicate the id of the cells. For cells
                with overlapping areas, the pixel is assigned for the cell with higher
                prediction scores.
        """
        preds = self.predict(image, **kwargs)

        return patches_to_label(
            preds,
            input_size=image.shape[:2],
            score_threshold=score_threshold,
        )

    def predict_on_large_image(
        self,
        image: ArrayLike,
        gs: int,
        ss: int,
        *,
        scaling: float = 1,
        min_area: int = 0,
        nms_iou: float = 0.3,
        segmentation_threshold: float = 0.5,
        score_threshold: float = 0.5,
        return_label: bool = False,
        **kwargs,
    ) -> Union[dict, ArrayLike]:
        """Make prediction on very large image by dividing into a grid.

        Direct model prediction on large image may cause out-of-memory error. This
        method divided the large image to smaller grid and then stitch the results
        to form the complete prediction.

        Args:
            image: An image with (H, W, C) format.
            gs: An int value. Grid size of the computation.
            ss: An int value of stepping size. Must be small than gs to produce valid
                results.

        Keyword Args:
            scaling: A image scaling factor.
            nms_iou: Optional iou threshold for the non-max-suppression post-processing.
                Default is 0, which disable non-max-suppression.
            min_area: Optional minimum area for the instance to be included in the results.
                Default is 0.
            segmentation_threshold: Default is 0.5.
            return_label: Whether to output full model prediction or segmentation
                label.

        Returns:
            Segmention label if output_label is True, otherwise a dict of full model
                predictions.
        """

        get_padding = lambda a, gs, ss: (a - 1) // ss * ss + gs - a

        h, w = image.shape[:2]
        padded_img = np.pad(
            image, [[0, get_padding(h, gs, ss)], [0, get_padding(w, gs, ss)], [0, 0]]
        )

        preds = []
        for y0 in range(0, h, ss):
            for x0 in range(0, w, ss):
                logging.info(f"Processing grid {y0}-{x0}")
                pred = self.predict(
                    padded_img[y0 : y0 + gs, x0 : x0 + gs],
                    scaling=scaling,
                    min_area=min_area,
                    **kwargs,
                )
                pred = _remove_edge_instances(
                    pred,
                    [gs, gs],
                    remove_top=y0 != 0,
                    remove_bottom=(y0 + gs) != h,
                    remove_left=x0 != 0,
                    remove_right=(x0 + gs) != w,
                )

                logging.info(f"Transfer result...")
                boolean_mask = np.asarray(
                    pred["instance_mask"].squeeze((1, 2))
                    & (pred["pred_scores"] >= score_threshold)
                )
                pred = dict(
                    pred_scores=np.array(pred["pred_scores"]),
                    pred_locations=np.array(
                        pred["pred_locations"] + jnp.asarray([y0, x0])
                    ),
                    pred_bboxes=np.array(
                        pred["pred_bboxes"] + jnp.asarray([y0, x0, y0, x0])
                    ),
                    pred_masks=np.array(
                        pred["instance_output"] >= segmentation_threshold
                    ),
                    pred_masks_y0=np.array(pred["instance_yc"][:, 0, 0] + y0),
                    pred_masks_x0=np.array(pred["instance_xc"][:, 0, 0] + x0),
                )
                pred = jax.tree_util.tree_map(lambda x: x[boolean_mask], pred)
                preds.append(pred)

        logging.info(f"Postprocessing...")
        preds = jax.tree_util.tree_map(lambda *x: np.concatenate(x), *preds)
        valid_locs = (
            (preds["pred_locations"] > 0).all(axis=1)
            & (preds["pred_locations"][:, 0] < h)
            & (preds["pred_locations"][:, 1] < w)
        )
        # valid_locs = valid_locs & (preds["pred_scores"] >= score_threshold)
        # valid_locs = valid_locs & (np.count_nonzero(preds["pred_masks"]) >= min_area)

        preds = jax.tree_util.tree_map(lambda x: x[valid_locs], preds)

        # sort based on scores
        asort = np.argsort(preds["pred_scores"])[::-1]
        preds = jax.tree_util.tree_map(lambda x: x[asort], preds)

        # nms
        if nms_iou > 0:
            logging.info(f"nms...")
            scores = preds["pred_scores"]
            boxes = preds["pred_bboxes"]
            _, _, selections = sorted_non_max_suppression(
                scores,
                boxes,
                -1,
                threshold=nms_iou,
                return_selection=True,
            )

            preds = jax.tree_util.tree_map(lambda x: x[selections], preds)

        if return_label:
            logging.info(f"Generating label...")

            ps = preds["pred_masks"].shape[-1]

            label = np.zeros([h + ps, w + ps], dtype=int)
            cnt = len(preds["pred_masks"])
            for mask, y0, x0 in zip(
                preds["pred_masks"][::-1],
                preds["pred_masks_y0"][::-1],
                preds["pred_masks_x0"][::-1],
            ):
                yc, xc = np.where(mask)
                label[(yc + y0 + ps // 2, xc + x0 + ps // 2)] = cnt
                cnt -= 1

            return label[ps // 2 : ps // 2 + h, ps // 2 : ps // 2 + w]

        else:
            return preds

    @property
    def module(self) -> nn.Module:
        return self.model[0]

    @property
    def params(self) -> Params:
        return self.model[1]

    @params.setter
    def params(self, new_params):
        self.model = self.module, new_params

    @property
    def detector(self) -> lacss.modules.Detector:
        return self.module.detector

    # FIXME Change the detector without first compile the mode will result
    # in error. This is a hidden gotcha. Need to setup warning to the user.
    @detector.setter
    def detector(self, new_detector):
        from ..modules import Detector

        cur_module = self.module

        if isinstance(new_detector, dict):
            new_detector = Detector(**new_detector)

        if isinstance(new_detector, Detector):
            self.model = (
                lacss.modules.Lacss(
                    self.module.backbone,
                    self.module.lpn,
                    new_detector,
                    self.module.segmentor,
                ),
                self.model[1],
            )
        else:
            raise ValueError(f"{new_detector} is not a Lacss Detector")

    def save(self, save_path) -> None:
        """Save the model by pickling.
            In the form of (module, weights).

        Args:
            save_path: Path to the pkl file
        """
        with open(save_path, "wb") as f:
            pickle.dump(self.model, f)



import struct
import sys
from pathlib import Path
from zipfile import ZipFile

import jax.numpy as jnp
import numpy as np
import typer

import lacss.deploy.lacss_pb2 as LacssMsg
from lacss.ops import patches_to_label, sorted_non_max_suppression

from .predict import Predictor

app = typer.Typer(pretty_exceptions_enable=False)

def read_input(st = sys.stdin.buffer):
    msg_size = st.read(4)
    msg_size = struct.unpack(">i", msg_size)[0]

    msg = LacssMsg.Input()
    msg.ParseFromString(st.read(msg_size))

    image = msg.image
    np_img = np.frombuffer(image.data, dtype=">f4").astype("float32")
    np_img = np_img.reshape(image.channel, image.width, image.height)
    np_img = np_img.transpose(2, 1, 0)

    return np_img, msg.settings

def img_to_msg(msg, img):
    img = np.ascontiguousarray(img, dtype=">i2") # match java format

    msg.height = img.shape[0]
    msg.width = img.shape[1]
    msg.data = img.tobytes()

    assert len(msg.data) == msg.height * msg.width * 2

def write_result(label, score, st = sys.stdout.buffer):
    if len(label.shape) != 2 :
        raise ValueError(f"Expect 2D array as label. Got array of {label.shape}")

    msg = LacssMsg.Result()
    img_to_msg(msg.score, (score * 1000).astype(np.int16))
    img_to_msg(msg.label, label)

    msg_size_bits = struct.pack(">i", msg.ByteSize())

    st.write(msg_size_bits)
    st.write(msg.SerializeToString())

@app.command()
def main(modelpath: Path):
    modelpath = str(modelpath)
    if "!" in modelpath: # in a jar
        jarfile, rscpath = modelpath.split("!")
        with ZipFile(jarfile) as jar:
            with jar.open(rscpath) as modelfile:
                import pickle
                modelpath = pickle.load(modelfile)

    model = Predictor(modelpath)
    model.detector.test_max_output = 512

    print(f"lacss_server: loaded model from {modelpath}", file=sys.stderr)

    # cnt = 0

    while True:
        img, settings = read_input()
        img = img - img.min()
        img = img / img.max()

        print(f"received image {img.shape}", file=sys.stderr)

        preds = model.predict(
            img, 
            min_area=settings.min_cell_area,
            remove_out_of_bound=settings.remove_out_of_bound,
            scaling=settings.scaling,
        )

        # boxed based nms
        # _, _, mask = sorted_non_max_suppression(
        #     preds["pred_scores"],
        #     preds["pred_bboxes"],
        #     -1, # max number of output
        #     threshold=settings.nms_iou,
        #     min_score=settings.detection_threshold,
        #     return_selection=True,
        # )

        label = np.asarray(patches_to_label(
            preds, 
            img.shape[:2],
            score_threshold=settings.detection_threshold,
            threshold=settings.segmentation_threshold,
        ).astype(int)).transpose()

        # score image
        score = np.asarray(preds["pred_scores"])[label]
        assert score.shape == label.shape

        # make label continous
        k = np.unique(label)
        v = np.arange(len(k))
        mapping_ar = np.zeros(k.max() + 1, dtype=np.uint16)
        mapping_ar[k] = v
        label = mapping_ar[label]

        write_result(label, score)

        # imageio.imwrite(f"p_{cnt}.tif", np.asarray(label))
        # cnt+=1

if __name__ == "__main__":
    app()



from __future__ import annotations

from functools import partial
from typing import Optional

import jax
import optax

from .common import binary_focal_crossentropy

jnp = jax.numpy

EPS = jnp.finfo("float32").eps

# def _binary_focal_crossentropy(pred, gt, gamma=2.0):
#     p_t = gt * pred + (1 - gt) * (1.0 - pred)
#     focal_factor = (1.0 - p_t) ** gamma

#     bce = -jnp.log(jnp.clip(p_t, EPS, 1.0))

#     return focal_factor * bce


def detection_loss(batch, prediction, *, gamma=2.0):
    """LPN detection loss"""

    preds = prediction

    scores = preds["lpn_scores"]
    gt_scores = preds["lpn_gt_scores"]

    score_loss = 0.0
    cnt = EPS
    for k in scores:
        score_loss += binary_focal_crossentropy(
            scores[k],
            gt_scores[k],
            gamma=gamma,
        ).sum()
        cnt += preds["lpn_scores"][k].size

    return score_loss / cnt


def localization_loss(batch, prediction, *, delta=1.0):
    """LPN localization loss"""
    preds = prediction

    regrs = preds["lpn_regressions"]
    gt_regrs = preds["lpn_gt_regressions"]
    gt_scores = preds["lpn_gt_scores"]

    regr_loss = 0.0
    cnt = 1e-8
    for k in regrs:
        # h = optax.l2_loss(regrs[k], gt_regrs[k])
        h = optax.huber_loss(regrs[k], gt_regrs[k], delta=delta).mean(
            axis=-1, keepdims=True
        )
        mask = gt_scores[k] > 0
        regr_loss += jnp.sum(h, where=mask)
        # cnt += jnp.count_nonzero(mask)
        cnt += gt_scores[k].size

    return regr_loss / (cnt + EPS)


def lpn_loss(batch, prediction, *, gamma=2.0, w1=1.0, w2=1.0):
    """LPN loss"""

    return (
        detection_loss(batch, prediction, gamma=gamma) * w1
        + localization_loss(batch, prediction) * w2
    )



from __future__ import annotations

import tensorflow as tf

from .generator import _crop_and_resize


def _init_boolean_mask(inputs):
    if "centroids" in inputs:
        return tf.ones([len(inputs["centroids"])], dtype=tf.bool)
    elif "bboxes" in inputs:
        return tf.ones([len(inputs["bboxes"])], dtype=tf.bool)
    else:
        return None


def _value_pair(x):
    try:
        a, b = x
    except:
        a = x
        b = x

    return a, b


def _image_size(img):
    H = tf.shape(img)[-3]
    W = tf.shape(img)[-2]

    return (H, W)


def resize(inputs: dict, *, target_size: tuple[int, int], p: float = 1.0) -> dict:
    """Resize image and labels

    Args:
        inputs: data from TF dataset. Affected elements are:

            * image
            * image_mask
            * centroids
            * bboxes

    Keyword Args:
        target_size: target size
        p: probability of applying transformation
    """

    def _resize(inputs):
        H, W = _image_size(inputs["image"])
        target_y, target_x = target_size

        scaling_y = tf.cast(target_y, tf.float32) / tf.cast(H, tf.float32)
        scaling_x = tf.cast(target_x, tf.float32) / tf.cast(W, tf.float32)

        inputs["image"] = tf.image.resize(
            inputs["image"],
            [target_y, target_x],
            antialias=True,
        )

        if "image_mask" in inputs:
            image_mask = (inputs["image_mask"][..., None],)  # need a channel dimension
            image_mask = tf.cast(
                tf.image.resize(
                    inputs["image_mask"][..., None],  # need a channel dimension
                    [target_y, target_x],
                )
                >= 0.5,
                inputs["image_mask"].dtype,
            )
            inputs["image_mask"] = tf.squeeze(image_mask, axis=-1)

        if "centroids" in inputs:
            inputs["centroids"] *= tf.cast([scaling_y, scaling_x], tf.float32)

        if "bboxes" in inputs:
            inputs["bboxes"] *= tf.cast(
                [scaling_y, scaling_x, scaling_y, scaling_x], tf.float32
            )

        return inputs

    if p < 1.0 and tf.random.uniform([]) >= p:
        return inputs

    else:
        return _resize(inputs)


def random_resize(
    inputs: dict,
    *,
    scaling: float | tuple[float, float],
    keep_aspect_ratio: bool = False,
    p: float = 1.0,
) -> dict:
    """Resize image by a random amount

    Args:
        inputs: data from TF dataset. Affected elements are:

            * image
            * image_mask
            * centroids

    Keyword Args:
        scaling: range of scale, e.g, [0.8, 1.5].
        keep_aspect_ratio: Whether to scale x/y the same amount.
        p: probability of applying transformation

    """

    if p < 1.0 and tf.random.uniform([]) >= p:
        return inputs

    else:
        min_scale, max_scale = scaling
        # if min_scale == max_scale and max_scale < 1:
        #     min_scale = 1.0 - min_scale
        #     max_scale = 1.0 + max_scale

        H, W = _image_size(inputs["image"])

        if keep_aspect_ratio:
            scaling_y = tf.random.uniform([], min_scale, max_scale)
            scaling_x = scaling_y
        else:
            scaling_y = tf.random.uniform([], min_scale, max_scale)
            scaling_x = tf.random.uniform([], min_scale, max_scale)

        target_h = tf.round(scaling_y * tf.cast(H, tf.float32))
        target_w = tf.round(scaling_x * tf.cast(W, tf.float32))

        return resize(inputs, target_size=(target_h, target_w))


def crop_to_roi(
    inputs: dict,
    *,
    roi: tuple[int, int, int, int],
    area_ratio_threshold: float = 1.0,
    p: float = 1.0,
) -> dict:
    """Crop image to bounding-box ROI

    Args:
        inputs: data from TF dataset. Affected elements are:

            * image
            * image_mask
            * centroids
            * bboxes
            * masks

    Keyword Args:
        roi: Rectangle roi in yxyx format
        area_ratio_threshold: remove instances if the bbox's relative remaining area is below this threshold
        p: probability of applying transformation

    """

    def _crop_to_roi(inputs):
        y0, x0, y1, x1 = roi
        roi_h = tf.cast(y1 - y0, tf.float32)
        roi_w = tf.cast(x1 - x0, tf.float32)

        inputs["image"] = inputs["image"][y0:y1, x0:x1, :]

        if "image_mask" in inputs:
            inputs["image_mask"] = inputs["image_mask"][y0:y1, x0:x1]

        boolean_mask = _init_boolean_mask(inputs)

        if boolean_mask is None:
            return inputs

        if "centroids" in inputs:
            ctrds = inputs["centroids"] - tf.cast([y0, x0], tf.float32)
            boolean_mask &= (
                tf.reduce_all(ctrds > 0, axis=-1)
                & (ctrds[:, 0] < roi_h)
                & (ctrds[:, 1] < roi_w)
            )

        if "bboxes" in inputs:
            bboxes = inputs["bboxes"] - tf.cast([y0, x0, y0, x0], tf.float32)
            areas = (bboxes[:, 2] - bboxes[:, 0]) * (bboxes[:, 3] - bboxes[:, 1])

            bboxes = tf.minimum(
                bboxes, tf.cast([roi_h, roi_w, roi_h, roi_w], tf.float32)
            )
            bboxes = tf.maximum(bboxes, 0)
            new_areas = (bboxes[:, 2] - bboxes[:, 0]) * (bboxes[:, 3] - bboxes[:, 1])

            boolean_mask &= (new_areas / areas) >= area_ratio_threshold

        # so that TF won't complain boolean_mask has no shape
        boolean_mask = tf.ensure_shape(boolean_mask, [None])

        if "centroids" in inputs:
            inputs["centroids"] = tf.boolean_mask(ctrds, boolean_mask)

        if "bboxes" in inputs:
            orig_bboxes = tf.boolean_mask(inputs["bboxes"], boolean_mask)
            inputs["bboxes"] = tf.boolean_mask(bboxes, boolean_mask)

            if "masks" in inputs:
                target_shape = inputs["masks"].shape[1:3]

                orig_y0 = orig_bboxes[:, 0] - tf.cast(y0, orig_bboxes.dtype)
                orig_x0 = orig_bboxes[:, 1] - tf.cast(x0, orig_bboxes.dtype)
                orig_h = orig_bboxes[:, 2] - orig_bboxes[:, 0]
                orig_w = orig_bboxes[:, 3] - orig_bboxes[:, 1]

                new_bbox = inputs["bboxes"] - tf.stack(
                    [orig_y0, orig_x0, orig_y0, orig_x0], axis=-1
                )
                new_bbox = new_bbox / tf.stack(
                    [orig_h, orig_w, orig_h, orig_w], axis=-1
                )
                new_bbox = new_bbox * tf.cast(
                    (target_shape + target_shape), new_bbox.dtype
                )

                inputs["masks"] = _crop_and_resize(
                    tf.boolean_mask(inputs["masks"], boolean_mask),
                    new_bbox,
                    target_shape,
                )

        return inputs

    if p < 1.0 and tf.random.uniform([]) >= p:
        return inputs

    else:
        return _crop_to_roi(inputs)


def random_crop(
    inputs: dict,
    *,
    target_size: tuple[int, int],
    area_ratio_threshold: float = 1.0,
    p: float = 1.0,
) -> dict:
    """Random crop to a set target size

    Args:
        inputs: data from TF dataset. Affected elements are:

            * image
            * image_mask
            * centroids
            * bboxes
            * masks

    Keyword Args:
        target_size: the target size
        area_ratio_threshold: remove instances if the bbox's relative remaining area is below this threshold
        p: probability of applying transformation
    """

    H, W = _image_size(inputs["image"])

    target_h, target_w = _value_pair(target_size)

    if p < 1.0 and tf.random.uniform([]) >= p:
        return inputs

    else:
        y0 = int(tf.random.uniform([]) * tf.cast(H - target_h, tf.float32))
        x0 = int(tf.random.uniform([]) * tf.cast(W - target_w, tf.float32))

        return crop_to_roi(
            inputs,
            roi=(y0, x0, y0 + target_h, x0 + target_w),
            area_ratio_threshold=area_ratio_threshold,
        )


def pad(
    inputs: dict,
    *,
    paddings: int | tuple[int, int],
    constant_values: float = 0,
    p: float = 1.0,
) -> dict:
    """Pad image and labels.

    Args:
        inputs: data from TF dataset. Affected elements are:

            * image
            * image_mask
            * centroids
            * bboxes

    Keyword Args:
        paddings: either a tuple or a single value. If latter, use the same padding for both x and y axis
        constant_values: the value to fill the padded area
        p: probability of applying transformation
    """

    def _pad(inputs):
        padding_y, padding_x = _value_pair(paddings)
        padding_y0, padding_y1 = _value_pair(padding_y)
        padding_x0, padding_x1 = _value_pair(padding_x)

        inputs["image"] = tf.pad(
            inputs["image"],
            [[padding_y0, padding_y1], [padding_x0, padding_x1], [0, 0]],
            constant_values=constant_values,
        )

        if "image_mask" in inputs:
            image_mask = inputs["image_mask"]
            if len(image_mask.shape) == 2:
                padding_values = [[padding_y0, padding_y1], [padding_x0, padding_x1]]
            else:
                padding_values = [
                    [0, 0],
                    [padding_y0, padding_y1],
                    [padding_x0, padding_x1],
                ]
            inputs["image_mask"] = tf.pad(
                inputs["image_mask"],
                padding_values,
                constant_values=constant_values,
            )

        if "centroids" in inputs:
            inputs["centroids"] += tf.cast([padding_y0, padding_x0], tf.float32)

        if "bboxes" in inputs:
            inputs["bboxes"] += tf.cast(
                [padding_y0, padding_x0, padding_y0, padding_x0], tf.float32
            )

        return inputs

    if p < 1.0 and tf.random.uniform([]) >= p:
        return inputs

    else:
        return _pad(inputs)


def pad_to_size(
    inputs: dict,
    *,
    target_size: tuple[int, int],
    constant_values: float = 0,
    p: float = 1.0,
) -> dict:
    """Pad image and labels to a target size. Padding is applied so that the orginal scene is centered in the output.

    Args:
        inputs: data from TF dataset. Affected elements are:

            * image
            * image_mask
            * centroids
            * bboxes

    Keyword Args:
        target_size: target image size
        constant_values: the value to fill the padded area
        p: probability of applying transformation
    """

    if p < 1.0 and tf.random.uniform([]) >= p:
        return inputs

    H, W = _image_size(inputs["image"])
    target_h, target_w = _value_pair(target_size)

    padding_h = target_h - H
    padding_h = [padding_h // 2, padding_h - padding_h // 2]
    padding_w = target_w - W
    padding_w = [padding_w // 2, padding_w - padding_w // 2]

    return pad(inputs, paddings=[padding_h, padding_w], constant_values=constant_values)


def random_crop_or_pad(
    inputs: dict,
    *,
    target_size: tuple[int, int],
    constant_values: float = 0,
    area_ratio_threshold: float = 1.0,
    p: float = 1.0,
) -> dict:
    """Random crop or pad image to specified target_size.

    Args:
        inputs: data from TF dataset. Affected elements are:

            * image
            * image_mask
            * centroids
            * bboxes
            * masks

    Keyword Args:
        target_size: target size
        constant_values: the value to fill the padded area
        area_ratio_threshold: remove instances if the bbox's relative remaining area is below this threshold
        p: probability of applying transformation
    """

    if p < 1.0 and tf.random.uniform([]) >= p:
        return inputs

    else:
        H, W = _image_size(inputs["image"])

        target_h, target_w = _value_pair(target_size)

        h = tf.cast(tf.maximum(H, target_h), H.dtype)
        w = tf.cast(tf.maximum(W, target_w), W.dtype)

        inputs = pad_to_size(
            inputs, target_size=(h, w), constant_values=constant_values
        )

        inputs = random_crop(
            inputs, target_size=target_size, area_ratio_threshold=area_ratio_threshold
        )

        return inputs


def flip_left_right(inputs: dict, *, p: float = 1.0) -> dict:
    """Flip image left-right

    Args:
        inputs: data from TF dataset. Affected elements are:

            * image
            * image_mask
            * centroids
            * bboxes

    Keyword Args:
        p: probability of applying transformation

    """

    def _flip_left_right(inputs):
        H, W = _image_size(inputs["image"])

        inputs["image"] = tf.image.flip_left_right(inputs["image"])

        if "image_mask" in inputs:
            image_mask = inputs["image_mask"][..., None]
            inputs["image_mask"] = tf.squeeze(
                tf.image.flip_left_right(image_mask),
                axis=-1,
            )

        if "centroids" in inputs:
            inputs["centroids"] = inputs["centroids"] * [1, -1] + [0, W]

        if "bboxes" in inputs:
            inputs["bboxes"] = inputs["bboxes"] * [1, -1, 1, -1] + [0, W, 0, W]

        # if "masks" in inputs:
        #     inputs["masks"] = inputs["masks"][:, :, ::-1, :]

        return inputs

    if p < 1.0 and tf.random.uniform([]) >= p:
        return inputs

    else:
        return _flip_left_right(inputs)


def flip_up_down(inputs: dict, *, p: float = 1.0) -> dict:
    """Flip image up-down

    Args:
        inputs: data from TF dataset. Affected elements are:

            * image
            * image_mask
            * centroids
            * bboxes

    Keyword Args:
        p: probability of applying transformation

    """

    def _flip_up_down(inputs):
        H, W = _image_size(inputs["image"])

        inputs["image"] = tf.image.flip_up_down(inputs["image"])

        if "image_mask" in inputs:
            image_mask = inputs["image_mask"][..., None]
            inputs["image_mask"] = tf.squeeze(
                tf.image.flip_up_down(image_mask),
                axis=-1,
            )

        if "centroids" in inputs:
            inputs["centroids"] = inputs["centroids"] * [-1, 1] + [H, 0]

        if "bboxes" in inputs:
            inputs["bboxes"] = inputs["bboxes"] * [-1, 1, -1, 1] + [H, 0, H, 0]

        # if "masks" in inputs:
        #     inputs["masks"] = inputs["masks"][:, ::-1, :, :]

        return inputs

    if p < 1.0 and tf.random.uniform([]) >= p:
        return inputs

    else:
        return _flip_up_down(inputs)
